# cheating at the Sierpinski zoom game

# https://cutelifebot.github.io/sierpinski/
# Demo: https://cutelifebot.github.io/sierpinski/

# The position of a lower edge on the screen is measured relative to a
# reference position (the initial position of the mouse); this is the error.

# The error is used to control the position of the mouse on the screen
# relative to the reference position. Ideally we could just set the mouse
# position to be the error -- the mouse would perfectly track the edge, and
# we'd win the game.

# Unfortunately there will be some offset between the measured error and the
# ideal mouse position, which will cause drift. It's this that makes the game
# hard! We can counteract this by scaling up the error slightly before applying
# it to the mouse position. This has the effect of over-compensating for the
# measured error, bringing the edge closer to the reference position. Because the 

# This scale is effective at keeping the edge at the same place on the screen,
# but it doesn't always keep it at the reference position. If we have some
# measurement offset o, and we're applying scale s, then the feedback will
# exactly compensate for the real error when the measured error e is such that:

#   e * s = e + o

# The measured offsets will drift towards then settle at this value of e.
# Unfortunately for a small s (which is desirable to avoid oscillation), e is
# large -- we ideally want it to be 0 to maximise the range of useful motion.
# It is therefore desirable to tweak the measured error so that it closely
# matches the real error.

# The scale applied to the error is controlled by error_scale, and the offset
# is controlled by error_offset. To tune these, set the error_scale to
# something large (say 1.1), and run the program. Wait for the error e printed on
# the console to settle, then add  e*(error_scale-1)  to error_offset.
# error_scale can then be decreased (reducing oscillation), and error_offset
# can be tweaked again if necessary.

# One further interesting detail is that the overshoot that we apply is quite
# small -- much less than a pixel for small errors. The mouse position is
# dithered so that we compensate for small errors before they become large.

# This was much more fun than playing manually!

error_scale = 1.02
error_offset = -2.07

###

import time
import numpy as np
import cv2
from mss import mss
from Xlib import display
# import antigravity

d = display.Display()
s = d.screen()
root = s.root

# wait for a mask key press before starting
while 1:
    ptr = root.query_pointer()
    if ptr.mask != 0:
        break
while 1:
    ptr = root.query_pointer()
    if ptr.mask == 0:
        break


ptr = root.query_pointer()
sample_rect = {'top': ptr.root_y - 100, 'left': ptr.root_x + 10, 'width': 15, 'height': 200}

sct = mss()

while 1:
    if root.query_pointer().mask != 0:
        break

    img = np.array(sct.grab(sample_rect))
    stripe = np.mean(img, axis=(1, 2))
    error = np.where(stripe.astype(float) < 253)[0][-1] - 100

    print(error)

    feedback = (float(error) + error_offset) * error_scale

    feedback_dither = int(np.floor(feedback + np.random.random()))

    root.warp_pointer(ptr.root_x, ptr.root_y + feedback_dither)
    d.sync()

    time.sleep(0.05)
